/*******************************************************************************
  MMJACKPLAY: A .wav file player that uses the Jack audio server

  begin		: May 3, 2007
  copyright	: (c) 2007 by Michael Minn
  email		: See www.michaelminn.com
  last revision	: 5/3/2007

  This program is free software; you can redistribute it and/or modify
  it under the terms of version 2 of the GNU General Public License as 
  published by the Free Software Foundation.

*******************************************************************************/

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
//#include <unistd.h>
//#include <stdarg.h>
//#include <string.h>
//#include <sys/poll.h>
//#include <sys/time.h>
//#include <sys/ioctl.h>
//#include <linux/soundcard.h>

#include <jack/jack.h>

#include "mmuwave_file_header.h"


class mmjackplay_client
{
	public:
	char *name;

	// files only
	int file_mode;
	int file_position;
	int file_descriptor;
	mmuwave_file_header file_header;

	jack_client_t *client;
	jack_port_t *input_port[2];	// left, right
	jack_port_t *output_port[2];

	mmjackplay_client *next;

	mmjackplay_client()
	{
		name = NULL;
		file_mode = 0;
		file_position = 0;
		file_descriptor = -1;

		client = NULL;
		input_port[0] = NULL;
		input_port[1] = NULL;
		output_port[0] = NULL;
		output_port[1] = NULL;
	}

	int file_read(int frames, float *left, float *right);
	int file_write(int frames, float *left, float *right);
};

int mmjackplay_client::file_read(int frames, float *left, float *right)
{
	if (!this || (file_descriptor < 0) || (frames <= 0) || !left || !right)
		return -EINVAL;

	int file_data_size = frames * file_header.bytes_per_sample;
	unsigned char *file_data = new unsigned char[file_data_size + 16];
	int data_read = read(file_descriptor, file_data, file_data_size);
	if (data_read >= 0)
		memset(file_data + data_read, 0, file_data_size - data_read);
	else
	{
		int status = -errno;
		fprintf(stderr, "File read() failure: %s\n", strerror(errno));
		return status;
	}

	//fprintf(stderr, "file_read(%d frames) = %d of %d bytes (%d buffer)\n", 
	//	frames, data_read, file_data_size, jack_get_buffer_size(client));

	for (int channel = 0; channel < 2; ++channel)
	{
		unsigned char *x = file_data + ((channel % file_header.channel_numbers) * file_header.bits_per_sample / 8);
		float *y =left;
		if (channel > 0)
			y = right;

		for (int frame = 0; frame < frames; ++frame, x += file_header.bytes_per_sample, ++y)
		{
			if (file_header.bits_per_sample == 16)
			{
				int z = *((short*) x);
				*y = z / 32768.0;
			}

			else if (file_header.bits_per_sample == 24)
			{
				int z = *((int*) x);
				if (z & 0x800000)
					z |= 0xff000000;
				else
					z &= 0x007fffff;
				*y = z / 8388608.0;
			}

			else // assume 8-bit
				*y = (((int)*x) - 128) / 128.0;
		}
	}

	if (file_data)
		delete[] file_data;

	file_position += file_data_size;

	return file_data_size;
}

int mmjackplay_client::file_write(int frames, float *left, float *right)
{
	if (!this || (file_descriptor < 0) || (frames <= 0) || !left || !right)
		return -EINVAL;

	int file_data_size = frames * file_header.bytes_per_sample;
	unsigned char *file_data = new unsigned char[file_data_size + 16];

	for (int channel = 0; channel < 2; ++channel)
	{
		unsigned char *x = file_data + ((channel % file_header.channel_numbers) * file_header.bits_per_sample / 8);
		float *y =left;
		if (channel > 0)
			y = right;

		for (int frame = 0; frame < frames; ++frame, x += file_header.bytes_per_sample, ++y)
		{
			if (file_header.bits_per_sample == 16)
			{
				int z = (int)(*y * 32768);
				x[0] = (unsigned char) (z & 0xff);
				x[1] = (unsigned char) ((z >> 8) & 0xff);
			}
			else if (file_header.bits_per_sample == 24)
			{
				int z = (int)(*y * 8388608);
				x[0] = (unsigned char) (z & 0xff);
				x[1] = (unsigned char) ((z >> 8) & 0xff);
				x[2] = (unsigned char) ((z >> 16) & 0xff);
			}

			else // assume 8-bit
				*x = (unsigned char) ((*y * 128) + 128);
		}
	}

	if (write(file_descriptor, file_data, file_data_size) < 0)
	{
		int status = -errno;
		fprintf(stderr, "File write() failure: %s\n", strerror(errno));
		return status;
	}

	if (file_data)
		delete[] file_data;

	file_position += file_data_size;

	return file_data_size;
}

static int interrupted = 0;

void player_interrupt(int signal)
{
	interrupted = 1;
}

void player_shutdown (void *arg)
{
	interrupted = 1;
}

void player_exit(mmjackplay_client *clients)
{
	for (mmjackplay_client *client = clients; client; client = client->next)
	{
		fprintf(stderr, "player_exit()\n");

		for (int side = 0; side < 2; ++side)
			if (client->output_port[side])
			{
				int buffer_frames = (int) jack_get_buffer_size(client->client);
				float *buffer = (float*) jack_port_get_buffer(client->output_port[side], buffer_frames);
				if (buffer)
				{
					fprintf(stderr, "clearing\n");
					memset(buffer, 0, buffer_frames * sizeof(float));
				}
			}

		if (client->client)
		{
			jack_transport_locate(client->client, 0);
			jack_transport_stop(client->client);
		}

		for (int side = 0; side < 2; ++side)
		{
			if (client->input_port[side])
				jack_port_disconnect(client->client, client->input_port[side]);

			if (client->output_port[side])
			{
				/**
				int buffer_frames = (int) jack_get_buffer_size(client->client);
				float *buffer = (float*) jack_port_get_buffer(client->output_port[side], buffer_frames);
				if (buffer)
				{
					fprintf(stderr, "clearing\n");
					memset(buffer, 0, buffer_frames * sizeof(float));
				}
				**/

				jack_port_disconnect(client->client, client->output_port[side]);
			}
		}

		if ((client != clients) && (client->file_descriptor >= 0))
			client->file_header.write_header(client->file_descriptor);

		if (client->file_descriptor >= 0)
			close(client->file_descriptor);

		if (client->client)
			jack_client_close (client->client);
	}

	exit (0);
}
#include <math.h>


int player_run(jack_nframes_t nframes, void *arg)
{
	mmjackplay_client *client = (mmjackplay_client*) arg;

	if (!client || (client->file_descriptor < 0))
		return -EINVAL;

	//jack_default_audio_sample_t
	float *input[2] = { NULL, NULL };
	float *output[2] = { NULL, NULL };
	input[0] = (float*) jack_port_get_buffer (client->input_port[0], nframes);
	input[1] = (float*) jack_port_get_buffer (client->input_port[1], nframes);
	output[0] = (float*) jack_port_get_buffer (client->output_port[0], nframes);
	output[1] = (float*) jack_port_get_buffer (client->output_port[1], nframes);

	if (!input[0] || !output[0])
	{
		fprintf(stderr, "jack_port_get_buffer() failure\n");
		return -1;
	}

	if (jack_transport_query(client->client, NULL) != JackTransportRolling)
	{
		memset(output[0], 0, nframes * sizeof(float));
		memset(output[1], 0, nframes * sizeof(float));
		return 0;
	}

	/**
	static int sine_length = 0;
	static float *sine = NULL;
	if (!sine || (sine_length != (int) nframes))
	{
		float multiple = 10;
		sine = new float[nframes];
		for (int frame = 0; frame < (int) nframes; ++frame)
			sine[frame] = sin(2.0 * M_PI * frame * multiple / (float) nframes);
	}

	memcpy(output[0], sine, nframes * sizeof(float));
	memcpy(output[1], sine, nframes * sizeof(float));
	**/

	/**
	float hz = 110;
	float rate = 44100;
	for (int frame = 0; frame < (int) nframes; ++frame, ++client->file_position)
	{
		output[0][frame] = sin(2.0 * M_PI * client->file_position * hz / rate);
		output[1][frame] = output[0][frame];
	}
	**/

	if (client->file_mode == O_RDONLY)
		client->file_read(nframes, output[0], output[1]);

	else
	{
		client->file_write(nframes, input[0], input[1]);
		memcpy(output[0], input[0], nframes * sizeof(float));
		memcpy(output[1], input[1], nframes * sizeof(float));
	}

	return 0;      
}

int player_seek(jack_transport_state_t state, jack_position_t *pos, void *arg)
{
	mmjackplay_client *client = (mmjackplay_client*) arg;

	if (!client || !pos || (client->file_descriptor <= 0))
		return -EINVAL;

	int position = sizeof(client->file_header) + pos->frame * (int) client->file_header.bytes_per_sample;

	//fprintf(stderr, "player_seek(%d frames = %d bytes)\n", pos->frame, position);

	if (lseek(client->file_descriptor, position, SEEK_SET) < 0)
	{
		int status = -errno;
		fprintf(stderr, "lseek(%s) failure: %s\n", client->name, strerror(errno));
		return status;
	}

	client->file_position = position;

	return 1;
}

void list_clients(mmjackplay_client *clients)
{
	fprintf(stderr, "===========================\n");
	for (mmjackplay_client *client = clients; client; client = client->next)
	{
		if (client->client)
			fprintf(stderr, "Client %s (%s)\n", client->name, jack_get_client_name(client->client));
		else
			fprintf(stderr, "Client: %s (not open)\n", client->name);

		if (client->file_descriptor >= 0)
			fprintf(stderr, "  File descriptor: %d\n", client->file_descriptor);

		if (client->input_port[0])
			fprintf(stderr, "  Inputs: %s, %s\n", jack_port_name(client->input_port[0]), 
				jack_port_name(client->input_port[1]));
		if (client->output_port[0])
			fprintf(stderr, "  Outputs: %s, %s\n", jack_port_name(client->output_port[0]), 
				jack_port_name(client->output_port[1]));
	}
	fprintf(stderr, "===========================\n");
}

int print_syntax()
{
	fprintf(stderr, "SYNTAX: mmjackplay <client> ... <client>\n");
	fprintf(stderr, "  <client> can be a JACK client (from jack_lsp)\n");
	fprintf(stderr, "   or a .wav file (must have .wav suffix)\n");
	return -EINVAL;
}

int is_wave_file(char *name)
{
	return name && (strlen(name) > 4) && !strcasecmp(name + strlen(name) - 4, ".wav");
}

int main(int argc, char **argv)
{
	if (argc < 3)
		return print_syntax();

	fprintf(stderr, "\x1b[2J\x1b[Hmmjackplay v2007.05.07 (c) 2007 by Michael Minn\n\n\n");

	// Build list of files/clients from command line
	mmjackplay_client *clients = NULL;
	for (int arg = 1; arg < argc; ++arg)
	{
		mmjackplay_client *client = new mmjackplay_client;
		client->name = argv[arg];

		mmjackplay_client **end = &clients;
		while (*end)
			end = &((*end)->next);
		*end = client;
	}

	// Open the files
	for (mmjackplay_client *client = clients; client; client = client->next)
		if (is_wave_file(client->name) && (client == clients))
		{
			// If the first argument is a file, it's the one and only input file
			client->file_mode = O_RDONLY;
			if ((client->file_descriptor = open(client->name, client->file_mode)) < 0)
			{
				fprintf(stderr, "Failure opening %s: %s\n", client->name, strerror(errno));
				return -1;
			}
			else if (read(client->file_descriptor, &client->file_header, sizeof(client->file_header)) < 0)
			{
				fprintf(stderr, "Failure reading file from %s: %s\n", client->name, strerror(errno));
				return -1;
			}
			else if (!client->file_header.is_valid_header())
			{
				fprintf(stderr, "File is not a valid WAVE file: %s\n", client->name);
				return -1;
			}
		}
		else if (is_wave_file(client->name))
		{
			client->file_mode = O_WRONLY | O_CREAT | O_TRUNC;
			if ((client->file_descriptor = open(client->name, client->file_mode, 0644)) < 0)
			{
				fprintf(stderr, "Failure opening %s: %s\n", client->name, strerror(errno));
				return -1;
			}
		}

	// Attach to any existing clients
	for (mmjackplay_client *client = clients; client; client = client->next)
		if (!is_wave_file(client->name))
		{
			//fprintf(stderr, "jack_client_open(%s)\n", client->name);
			jack_status_t status = (JackStatus) 0;
			jack_options_t options = JackNullOption;
			if ((client->client = jack_client_open (client->name, options, &status, NULL)) == NULL)
			{
				fprintf(stderr, "Unable to connect to client %s through JACK server: code %d\n",
					client->name, (int) status);
				player_exit(clients);
			}

			else if (status == JackNameNotUnique)
				fprintf(stderr, "jack_client_open(%s) non-unique client name replaced with %s\n",
					client->name, jack_get_client_name(client->client));

			else if ((int) status != 0)
			{
				fprintf(stderr, "jack_client_open(%s) failure: status 0x%x\n", client->name, (int) status);
				player_exit(clients);
			}
		}

	//list_clients(clients);

	// Create clients and ports for files
	for (mmjackplay_client *client = clients; client; client = client->next)
		if (is_wave_file(client->name))
		{
			//fprintf(stderr, "jack_client_new(%s)\n", client->name);
			client->client = jack_client_new(client->name);
			if (!client->client)
			{
				fprintf (stderr, "jack_client_new(%s): Unable to connect to JACK server\n", client->name);
				player_exit(clients);
			}

			// Verify matching sample rate
			if (client->file_header.sample_rate != jack_get_sample_rate(client->client))
			{
				fprintf(stderr, "File sample rate %d does not match JACK server rate %d\n", 
					client->file_header.sample_rate, jack_get_sample_rate(client->client));
				//player_exit(clients);
			}

			// Create ports
			client->input_port[0] = jack_port_register(client->client, "input_l", 
				JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 4096);
			client->input_port[1] = jack_port_register(client->client, "input_r", 
				JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 4096);
			client->output_port[0] = jack_port_register(client->client, "output_l", 
				JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 4096);
			client->output_port[1] = jack_port_register(client->client, "output_r", 
				JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 4096);

			// Define callbacks
			jack_set_sync_callback(client->client, player_seek, client);
			jack_set_process_callback(client->client, player_run, client);
			jack_on_shutdown(client->client, player_shutdown, client);

			// Activate the client - must be done before connecting ports
			int activate_status = jack_activate(client->client);
			if (activate_status < 0)
			{
				fprintf(stderr, "jack_activate() failed: code %d\n", activate_status);
				player_exit(clients);
			}
		}

	// Find ports for non-file clients
	for (mmjackplay_client *client = clients; client; client = client->next)
		if (!is_wave_file(client->name))
		{
			const char **ports = jack_get_ports(client->client, client->name, NULL, JackPortIsInput);
			if (ports)
			{
				client->input_port[0] = jack_port_by_name(client->client, ports[0]);
				if (ports[1])
					client->input_port[1] = jack_port_by_name(client->client, ports[1]);
				else
					client->input_port[1] = client->input_port[0];

				free (ports);
			}
			
			ports = jack_get_ports(client->client, client->name, NULL, JackPortIsOutput);
			if (!ports)
			{
				fprintf(stderr, "jack_get_ports(%s, Output) failed\n", client->name);
				player_exit(clients);
			}

			client->output_port[0] = jack_port_by_name(client->client, ports[0]);
			if (ports[1])
				client->output_port[1] = jack_port_by_name(client->client, ports[1]);
			else
				client->output_port[1] = client->output_port[0];

			free (ports);
		}

	// Connect ports between clients in a daisy chain
	for (mmjackplay_client *client = clients; client && client->next; client = client->next)
		for (int side = 0; side < 2; ++side)
		{
			//fprintf(stderr, "Connecting %s (%s) -> %s (%s)\n", 
			//	client->name, jack_port_name(client->output_port[side]), 
			//	client->next->name, jack_port_name(client->next->input_port[side]));

			int status = jack_connect (client->client, jack_port_name(client->output_port[side]), 
						jack_port_name(client->next->input_port[side]));

			if (status != 0)
			{
				fprintf(stderr, "jack_connect(%s, %s) %s -> %s failed: %s (%d)\n",
					jack_port_name(client->output_port[side]), 
					jack_port_name(client->next->input_port[side]),
					client->name, client->next->name, strerror(status), status);
				player_exit(clients);
			}
		}

	// Diagnostic list
	list_clients(clients);

	signal(SIGINT, player_interrupt);

	jack_transport_locate(clients->client, 0);
	jack_transport_start(clients->client);

	while (!interrupted)
	{
		sleep(1);
		int seconds = (clients->file_position / clients->file_header.bytes_per_second) % 60;
		int minutes = (clients->file_position / clients->file_header.bytes_per_second) / 60;
		char text_buffer[256];
		sprintf(text_buffer, "\x1b[2H%03d:%02d", minutes, seconds);
		write(1, text_buffer, strlen(text_buffer));
	}

	player_exit(clients);

	return 0;
}
	/**
	jack_status_t status = (JackStatus) 0;
	jack_options_t options = JackNullOption;
	char *client_name = "mmjackplay";
	player.client = jack_client_open (client_name, options, &status, NULL);

	if (!player.client || (status & JackServerFailed))
	{
		fprintf (stderr, "jack_client_open(): Unable to connect to JACK server\n");
		return -EINVAL;
	}

	if (status & JackNameNotUnique) 
	{
		fprintf(stderr, "jack_client_open() failure: client name %s not unique\n", client_name);
		return -EINVAL;
	}

	if (status & JackServerStarted) 
		fprintf (stderr, "JACK server started\n");
	**/

